from functools import partial

__all__ = ['pipe']


class Pipe(object):
    """Provide simple syntax for function composition, such as
    f[g(x)], f(g(h(x))) etc.(through << and >> operators)
    Usage example:
    >>> func = Pipe() << (lambda x: x + 1) << (lambda x: x**2)
    >>> print(func(5))
    26
    >>> func = Pipe() >> (lambda x: x + 1) >> (lambda x: x**2)
    >>> print(func(5))
    36
    >>> func = Pipe() >> (lambda x: filter(lambda y: y < 6, x)) >> sum
    >>> print(func(range(10)))
    15
    >>> func = Pipe(lambda l: map(lambda x: x*x, l)) >> sum >> (lambda x: x**(1/2))
    >>> print(func(range(3, 5)))
    5.0
    """
    __slots__ = 'f'

    def __init__(self, f=lambda x: x, *args, **kwargs):
        self.f = partial(f, *args, **kwargs) if any([args, kwargs]) else f

    def __str__(self):
        return "<class 'Pipe'>"

    @classmethod
    def __compound(cls, f, g):
        return cls(lambda *args, **kwargs: f(g(*args, **kwargs)))

    def __rshift__(self, g):
        return self.__class__.__compound(g, self.f)

    def __lshift__(self, g):
        return self.__class__.__compound(self.f, g)

    def __rrshift__(self, g):
        return self.__class__() >> g >> self.f

    def __rlshift__(self, g):
        return self.__class__() >> self.f >> g

    def __le__(self, other):
        return self.__call__(other)

    def __ror__(self, other):
        return self.__call__(other)

    def __or__(self, other):
        return self.__call__(other)

    def __call__(self, *args, **kwargs):
        if not args:
            return self.__class__() >> (lambda x: x)
        if callable(args[0]):
            return self.__class__() >> args[0]
        return self.f(*args, **kwargs)


pipe = Pipe()

if __name__ == '__main__':
    pass



from functools import partial, wraps
from inspect import getfullargspec


def currying(func):
    """A decorator that makes the function curried
    Usage example:
    >>> @currying
    ... def sum5(a, b, c, d, e):
    ...     return a + b + c + d + e
    ...
    >>> sum5(1)(2)(3)(4)(5)
    15
    >>> sum5(1, 2, 3)(4, 5)
    15
    """

    @wraps(func)
    def _currying(*args, **kwargs):
        f = func
        count = 0
        while isinstance(f, partial):
            if f.args:
                count += len(f.args)
            f = f.func

        spec = getfullargspec(f)

        if count == len(spec.args) - len(args):
            return func(*args, **kwargs)

        return currying(partial(func, *args, **kwargs))

    return _currying




from .currying import currying
from .tailrecurve import tail_recurse_optimizer
from .infix import infix, is_a, to, step, has, take, drop

__all__ = [
    'currying',
    'tail_recurse_optimizer',
    'infix',
    'is_a',
    'to',
    'step',
    'has',
    'take',
    'drop'
]


import inspect
import functools

__all__ = ['tail_recurse_optimizer']


class TailRecurseException(BaseException):
    def __init__(self, args, kwargs):
        self.args = args
        self.kwargs = kwargs


def tail_recurse_optimizer(func):
    """
    A decorator for tail recursion optimization.
    Usage example:
    >>> @tail_recurse_optimizer
    ... def fib(i, a=0, b=1):
    ...     if i == 0:
    ...         return a
    ...     else:
    ...         return fib(i - 1, b, a + b)
    >>> fib(1000)
    43466557686937456435688527675040625802564660517371780402481729089536555417949051890403879840079255169295922593080322634775209689623239873322471161642996440906533187938298969649928516003704476137795166849228875
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        currentframe = inspect.currentframe()  # sys._current_frames() 也可
        if currentframe.f_back and currentframe.f_back.f_back \
                and currentframe.f_back.f_back.f_code == currentframe.f_code:
            raise TailRecurseException(args, kwargs)
        else:
            while 1:
                try:
                    return func(*args, **kwargs)
                except TailRecurseException as e:
                    args = e.args
                    kwargs = e.kwargs

    return wrapper



from functools import partial

__all__ = [
    'infix',
    'is_a',  # infix(isinstance)
    'to',
    'step',
    'has',  # infix(hasattr)
    'take',
    'drop',
]


class infix(partial):
    """A decorator that allows functions to be used as infix functions.
    Usage example:
    >>> @infix
    ... def plus(a, b):
    ...     return a + b
    ...
    >>> 40 /plus/ 2
    42
    >>> 5 /plus/ 6
    11
    >>> list(1 /to/ 4 /step/ -1)
    [3, 2, 1]
    >>> def pow(a, b):
    ...     return a ** b
    ...
    >>> pow = infix(pow)
    >>> 3 /pow/ 2
    9
    >>> 3 /is_a/ int
    True
    """

    def __truediv__(self, right):
        return self(right)

    def __rtruediv__(self, left):
        return infix(self.func, left)


is_a = infix(isinstance)
has = infix(hasattr)


@infix
def to(start, end):
    return range(start, end)


@infix
def step(obj, step):
    return obj[::step]


@infix
def take(obj, n):
    return obj[:n]


@infix
def drop(obj, n):
    return obj[n:]



"""
anoyfunc(anonymous function) makes the function definition more wonderful.

Usage example:
def f(x):
    return x**(1/2)

def g(x, y):
    return x + 2*y

def h(x):
    return x**2 + x - 1

f or lambda x: x**(1/2) is equivalent to _**(1/2) or __**(1/2)
g or lambda x, y: x + 2*y is equivalent to _ + 2*_ or __ + 2*_
h or lambda x: x**2 + x - 1 is equivalent to __**2 + __ - 1

when you use this ,just write two lines of code:
from anoyfunc import underscore as _
from anoyfunc import uunderscore as __

So amazing!

Enjoy it!
"""
import random
import string
from itertools import repeat, count

__all__ = ['underscore', 'uunderscore']


def random_name():
    return "".join(random.choice(string.ascii_letters) for _ in range(14))


def fmap(format):
    def applyier(self, other):
        fmt = format.replace("self", self._format)
        if other.__class__ == self.__class__:
            return self.__class__(fmt.replace("other", other._format),
                                  dict(list(self._format_args.items()) + list(other._format_args.items())))
        elif isinstance(other, (Underscore, UUnderscore)):
            return UUnderscore(fmt.replace("other", other._format),
                               dict(list(self._format_args.items()) + list(other._format_args.items())))
        else:
            name = random_name()
            return self.__class__(fmt.replace("other", "%%(%s)r" % name),
                                  dict(list(self._format_args.items()) + [(name, other)]))

    return applyier


def unary_fmap(format):
    def applyier(self):
        fmt = "(%s)" % format.replace("self", self._format)
        return self.__class__(fmt,
                              self._format_args)

    return applyier


class Underscore(object):
    __slots__ = "_format", "_format_args", "_func_from_format"

    def __init__(self, format_="_", format_args=None):
        self._format = format_
        self._format_args = format_args or {}
        self._func_from_format = eval(self._create_func_from_format())

    def __str__(self):
        str_func = self._create_func_from_format()
        return "<class 'Underscore':a callable object roughly equivalent to function: %s>" % str_func

    def _create_func_from_format(self):
        args = map("".join, zip(repeat("x"), map(str, count(1))))
        l, r = [], self._format
        while r.count("_"):
            n = next(args)
            r = r.replace("_", n, 1)
            l.append(n)
        r = r % self._format_args
        return "lambda {left}: {right}".format(left=", ".join(l), right=r)

    def __getitem__(self, k):
        return self.__class__(self._format + '[%s]' % k, self._format_args)

    def __getattr__(self, name):
        return self.__class__(self._format + '.%s' % name,
                              self._format_args)

    def __call__(self, *args):
        return self._func_from_format(*args)

    __add__ = fmap("self + other")
    __sub__ = fmap("self - other")
    __mul__ = fmap("self * other")
    __truediv__ = fmap("self / other")
    __floordiv__ = fmap("self // other")
    __mod__ = fmap("self %% other")
    __pow__ = fmap("self ** other")
    __neg__ = unary_fmap("-self")
    __pos__ = unary_fmap("+self")
    __invert__ = unary_fmap("~self")
    __lshift__ = fmap("self << other")
    __rshift__ = fmap("self >> other")
    __and__ = fmap("self & other")
    __or__ = fmap("self | other")
    __xor__ = fmap("self ^ other")
    __radd__ = fmap("other + self")
    __rsub__ = fmap("other - self")
    __rmul__ = fmap("other * self")
    __rtruediv__ = fmap("other / self")
    __rfloordiv__ = fmap("other // self")
    __rmod__ = fmap("other %% self")
    __rpow__ = fmap("other ** self")
    __rlshift__ = fmap("other << self")
    __rrshift__ = fmap("other >> self")
    __rand__ = fmap("other & self")
    __ror__ = fmap("other | self")
    __rxor__ = fmap("other ^ self")
    __lt__ = fmap("self < other")
    __le__ = fmap("self <= other")
    __gt__ = fmap("self > other")
    __ge__ = fmap("self >= other")
    __eq__ = fmap("self == other")
    __ne__ = fmap("self != other")


class UUnderscore(Underscore):

    def __init__(self, format_="__", format_args=None):
        self._format = format_
        self._format_args = format_args or {}
        self._func_from_format = eval(self._create_func_from_format())

    def __str__(self):
        str_func = self._create_func_from_format()
        return "<class 'anoyfunc.UUnderscore':a callable object roughly equivalent to function: %s>" % str_func

    def _create_func_from_format(self):
        r = self._format.replace("__", 'x')
        args = map("".join, zip(repeat("x"), map(str, count(1))))
        l = []
        b = False
        while r.count("_"):
            b = True
            n = next(args)
            r = r.replace("_", n, 1)
            l.append(n)
        r = r % self._format_args
        if not b:
            return "lambda x: {r}".format(r=r)
        return "lambda x, {left}: {right}".format(left=", ".join(l), right=r)


underscore = Underscore()
uunderscore = UUnderscore()

if __name__ == '__main__':
    pass


from erlei.underscore import underscore as _
from erlei.underscore import uunderscore as __
from erlei.pipe import pipe

__all__ = [
    '_',
    '__',
    'pipe'
]

__version__ = "1.0.0"
name = "erlei"